#!/usr/bin/env python
"""Test suite for py:mod:`plastid.util.services.misc`"""
import numpy
import unittest
from nose.plugins.attrib import attr
from plastid.util.services.misc import guess_formatter, number


@attr(test="unit")
class TestMisc(unittest.TestCase):
    def setUp(self):
        self.tests = [
            ("nan", numpy.nan),
            ("Nan", numpy.nan),
            ("NaN", numpy.nan),
            ("None", numpy.nan),
            ("none", numpy.nan),
            ("Inf", numpy.inf),
            ("inf", numpy.inf),
            ("-Inf", -numpy.inf),
            ("-inf", -numpy.inf),
            ("5", 5),
            ("5.0", 5),
            ("-5", -5),
            ("-5.0", -5),
            ("0", 0),
            ("5.1", 5.1),
            ("-5.1", -5.1),
            ("a5", "a5"),
            ("5a", "5a"),
            ("-5.1a", "-5.1a"),
            ("some_string", "some_string"),
            ("True", True),
            ("False", False),
            ("true", True),
            ("false", False),
            ("True5", "True5"),
            ("5True", "5True"),
            ("1e10", 1e10),
            ("1e-10", 1e-10),
            ("-1e10", -1e10),
            ("-1e-10", -1e-10),
        ]

    def test_guess_formatter(self):
        """Test parsing of strings to floats, ints, nans, infs, bools, and strings"""
        for source, dest in self.tests:
            if isinstance(dest, float):
                if not numpy.isnan(dest):
                    self.assertEquals(guess_formatter(source), dest)
                else:
                    self.assertTrue(numpy.isnan(guess_formatter(source)))
            else:  #if isinstance(dest,int):
                self.assertEquals(guess_formatter(source), dest)

    def test_number(self):
        """Test numerical parsing of strings to floats, ints, nans, and infs"""
        for source, dest in self.tests:
            if isinstance(dest, bool):
                pass
            elif isinstance(dest, float):
                if not numpy.isnan(dest):
                    self.assertEquals(number(source), dest)
                else:
                    self.assertTrue(numpy.isnan(number(source)))
            elif isinstance(dest, int):
                self.assertEquals(number(source), dest)
            else:
                self.assertRaises(ValueError, number, source)



#!/usr/bin/env python
"""Various wrappers and utilities for opening, closing, and writing files.

Important methods
-----------------
:py:func:`argsopener`
    Opens a file for writing within a command-line script and writes to it
    all command-line arguments as a pretty-printed dictionary of metadata,
    commented out. The open file handle is then returned for subsequent
    writing

:py:func:`read_pl_table`
    Wrapper function to open a table saved by one of :data:`plastid`'s
    command-line scripts into a :class:`pandas.DataFrame`.

:py:func:`opener`
    Guesses whether a file is bzipped, gzipped, zipped, or uncompressed based upon 
    file extension, opens it appropriately, and returns a file-like object.

:py:func:`NullWriter`
    Returns an open filehandle to the system's null location.
"""
import sys
import os
import pandas as pd
from plastid.util.io.filters import AbstractWriter
from plastid.util.services.mini2to3 import Iterable


class NullWriter(AbstractWriter):
    """Writes to system-dependent null location.
    On Unix-like systems & OSX, this is typically /dev/null. On Windows, simply "nul"
    """

    def __init__(self):
        self.stream = open(os.devnull, "w")

    def filter(self, stream):
        return stream

    def __repr__(self):
        # unusual repr, but useful for documentation by Sphinx
        return "NullWriter()"

    def __str__(self):
        return self.__repr__()


def multiopen(inp, fn=None, args=None, kwargs=None):
    """Normalize filename/file-like/list of filename or file-like to a list of appropriate objects
    
    If not list-like, `inp` is converted to a list. Then, for each element `x` in
    `inp`, if `x` is file-like, it is yielded. Otherwise, `fn` is applied to `x`,
    and the result yielded. 
    
    
    Parameters
    ----------
    inp : str, file-like, or list-like of either of those
        Input describing file(s) to open
    
    fn : callable, optional
        Callable to apply to input to open it
    
    args : tuple, optional
        Tuple of positional arguments to pass to `fn`
        
    kwargs : keyword arguments
        Arguments to pass to `fn`
        
    Yields
    ------
    Object
        Result of applying `fn` to filename(s) in `inp` 
    """
    if fn is None:
        fn = lambda x, **z: x

    if args is None:
        args = ()

    if kwargs is None:
        kwargs = {}

    if isinstance(inp, str):
        out = [inp]
    elif isinstance(inp, Iterable):
        out = inp

    for obj in out:
        if isinstance(obj, str):
            yield fn(obj, *args, **kwargs)
        else:
            yield obj


def opener(filename, mode="r", **kwargs):
    """Open a file, detecting whether it is compressed or not, based upon
    its file extension. Extensions are tested in the following order:
    
       +----------------+------------------+
       | File ends with | Presumed to be   |
       +================+==================+
       | gz             |    gzipped       |
       +----------------+------------------+
       | bz2            |    bzipped       |
       +----------------+------------------+
       | zip            |    zipped        |
       +----------------+------------------+
       | anything else  |    uncompressed  |
       +----------------+------------------+
        
    Parameters
    ----------
    filename : str
        Name of file to open
    
    mode : str
        Mode in which to open file. See Python standard
        libarary documentation on file opening modes for
        choices (e.g. "r", "a, "w" with or without "b")
    
    **kwargs
        Other parameters to pass to appropriate file opener 
    """
    if filename.endswith(".gz"):
        import gzip
        if "b" not in mode:
            mode += "b"
        call_func = gzip.GzipFile
    elif filename.endswith(".bz2"):
        import bz2
        if "b" not in mode:
            mode += "b"
        call_func = bz2.BZ2File
    elif filename.endswith(".zip"):
        import zipfile
        if "b" not in mode:
            mode += "b"
        call_func = zipfile.ZipFile
    else:
        call_func = open

    return call_func(filename, mode, **kwargs)


# TODO: needs unittest
def read_pl_table(filename, **kwargs):
    """Open a table saved by one of :data:`plastid`'s command-line scripts,
    passing default arguments to :func:`pandas.read_table`:
    
        ==========   =======
        Key          Value
        ----------   -------
        sep          `"\t"`
        comment      `"#"`
        index_col    `None`
        header       `0`
        ==========   =======
    
    Parameters
    ----------
    filename : str
        Name of file. Can be gzipped, bzipped, or zipped.
        
    kwargs : keyword arguments
        Other keyword arguments to pass to :func:`pandas.read_table`.
        Will override defaults.
        
    Returns
    -------
    :class:`pandas.DataFrame`
        Table of results
    """
    args = {
        "sep"       : "\t",
        "comment"   : "#",
        "index_col" : None,
        "header"    : 0,
    } # yapf: disable
    args.update(kwargs)
    table = pd.read_table(filename, **args)
    return table


# TODO: needs unit test
def write_pl_table(df, filename, sep="\t", header=True, index=None, **kwargs):
    """Wrapper function to write DataFrame `df` to a tab-delimited table, with header

    Parameters
    ----------
    df : :class:`~pandas.DataFrame`
        DataFrame to save

    filename : str
        Name of file to create

    **kwargs : keyword arguments, optional
        Any keyword argument readable by :meth:`pandas.DataFrame.to_csv`. 
    """
    return df.to_csv(filename, sep=sep, header=header, index=index, **kwargs)


def get_short_name(inpt, separator=os.path.sep, terminator=""):
    """Gives the basename of a filename or module name passed as a string.
    If the string doesn't match the pattern specified by the separator
    and terminator, it is returned unchanged. 
    
    Examples
    --------
    >>> get_short_name("test")
    'test'

    >>> get_short_name("test.py", terminator=".py")
    'test'

    >>> get_short_name("/home/jdoe/test.py", terminator=".py")
    'test'

    >>> get_short_name("/home/jdoe/test.py.py", terminator=".py")
    'test.py'

    >>> get_short_name("/home/jdoe/test.py.2")
    'test.py.2'
    
    >>> get_short_name("/home/jdoe/test.py.2", terminator=".py")
    'test.py.2'

    >>> get_short_name("plastid.bin.test", separator="\.", terminator="")
    'test'
    
    Parameters
    ----------
    inpt : str
        Input
    
    terminator : str
        File terminator (default: "")
    
    Returns
    -------
    str
    """
    import re
    tlen = len(terminator)
    if inpt[-tlen:] == terminator:
        inpt = inpt[:-tlen]
    try:
        stmp = re.split(separator, inpt)[-1]
    except AttributeError:
        return inpt

    return(stmp)

def argsopener(filename, namespace, mode="w", **kwargs):
    """Open a file for writing, and write to it command-line arguments
    formatted as a pretty-printed dictionary in comment metadata.
    
    Parameters
    ----------
    filename : str
        Name of file to open. If it terminates in `'.gz'` or `'.bz2'`
        the filehandle will write to a gzipped or bzipped file
                          
    namespace : :py:class:`argparse.Namespace`
        Namespace object from argparse.ArgumentParser
    
    mode : str
        Mode of writing (`'w'` or `'wb'`)
    
    **kwargs
        Other keyword arguments to pass to file opener
    
    Returns
    -------
    open filehandle
    """
    if "w" not in mode:
        mode += "w"
    fout = opener(filename, mode, **kwargs)
    fout.write(args_to_comment(namespace))
    return fout


def args_to_comment(namespace):
    """Formats a :class:`argparse.Namespace` into a comment block
    useful for printing in headers of output files
    
    Parameters
    ----------
    namespace  : :py:class:`argparse.Namespace`
        Namespace object returned by argparse.ArgumentParser
        
    Returns
    -------
    string
    """
    import datetime
    dtmp = namespace.__dict__
    ltmp = ["## date = '%s'" % datetime.datetime.today(), "## execstr = '%s'" % " ".join(sys.argv)]
    ltmp.append("## args = {  ")
    ltmp2 = pretty_print_dict(dtmp).split("\n")[1:-2]
    for i in range(len(ltmp2)):
        ltmp2[i] = "##" + ltmp2[i]
    sout = "\n".join(ltmp) + "\n" + ",\n".join(ltmp2) + "\n##        }\n"
    return sout


def pretty_print_dict(dtmp):
    """Pretty prints an un-nested dictionary 

    Parameters
    ----------
    dtmp : dict
    
    Returns
    -------
    str
        pretty-printed dictionary
    """
    ltmp = []
    keys = dtmp.keys()
    maxlen = 2 + max([len(K) for K in keys])
    for k, v in sorted(dtmp.items(), key=lambda x: x[0]):
        if type(v) == type(""):
            v = "'%s'" % v
        new_k = "'%s'" % k
        stmp = ("          {0:<%s} : {1}," % maxlen).format(new_k, v)
        ltmp.append(stmp)

    sout = "\n".join(ltmp)
    return "{\n%s\n}\n" % sout



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.gff_parent_types`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import REF_FILES
from plastid.bin.gff_parent_types import main
from plastid.util.services.decorators import catch_stderr

#===============================================================================
# INDEX: global constants used by tests
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.gff_parent_types",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="test_gff_parent_types"),
}

# Define tests as tuples of:
# 1. Command-line style arguments to pass to ``main``
# 2. A list of reference files that output should be compared against
# 3. A list of output files created by running ``main`` with the arguments provided in (1)
# 4. A list of strings specifying how equality should be evaluated
_outfile = os.path.join(test_info["temp_file_path"], "test_parent_types.txt")
gff_parent_types_tests = [
    # test ot ignoring any feature types
    (
        "%s %s" % (REF_FILES["yeast_gff"], _outfile), [REF_FILES["yeast_parent_child"]],
        [_outfile], [""]
    ),
    # test ignoring manu feature types
    (
        "%s %s --exclude ARS tRNA LTR_retrotransposon StopFeature X_element" %
        (REF_FILES["yeast_gff"], _outfile), [REF_FILES["yeast_parent_child_exclude_cols"]],
        [_outfile], [""]
    ),
]


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for :py:mod`plastid.bin.gff_parent_types`"""
    for x in execute_helper(test_info, gff_parent_types_tests):
        yield x



#!/usr/bin/env python
""":data:`plastid` includes the following command-line scripts. Click on a name below for more information:

    =========================   =============================================================================
    **Analysis of sequencing or quantitative data**                 
    ---------------------------------------------------------------------------------------------------------
    |counts_in_region|           Count the number of :term:`read alignments <alignment>` covering
                                 arbitrary regions of interest in the genome, and calculate
                                 read densities (in reads per nucleotide and in :term:`RPKM`)
                                 over these regions
    
    |cs|                         Count the number of :term:`read alignments<alignment>`
                                 and calculate read densities (in :term:`RPKM`)
                                 specifically for genes and sub-regions (5' UTR,
                                 CDS, 3' UTR)

    |get_count_vectors|          Fetch vectors of :term:`counts` or other quantitative data
                                 at each nucleotide position
                                 in one or more regions of interest, saving each vector
                                 as its own line-delimited text file
                             
    |make_wiggle|                Create `wiggle`_ or `bedGraph`_ files from alignment files
                                 after applying a read :term:`mapping rule` (e.g.
                                 to map :term:`ribosome-protected footprints <footprint>`
                                 at their :term:`P-sites <P-site offset>`), for
                                 visualization in a :term:`genome browser`
                             
    |metagene|                   Compute a :term:`metagene` profile of :term:`read alignments`,
                                 :term:`counts`, or quantitative data over one or more regions of interest,
                                 optionally applying a :term:`mapping rule`
                             
    |phase_by_size|              Estimate :term:`sub-codon phasing` in
                                 :term:`ribosome profiling` data
    
    |psite|                      Estimate position of ribosomal P-site within
                                 :term:`ribosome profiling` :term:`read alignments`
                                 as a function of read length
    -------------------------   -----------------------------------------------------------------------------
    **Generating or modifying genome annotations**    
    ---------------------------------------------------------------------------------------------------------
    |crossmap|                   Generate a :term:`mask file` annotating regions of the genome
                                 that fail to produce
                                 uniquely mapping reads under various alignment
                                 criteria, so that they may be excluded from future analyses
                             
    |gff_parent_types|           Examine parent-child relationships of features
                                 in a `GFF3`_ file
                             
    |reformat_transcripts|       Convert transcripts between `BED`_, `BigBed`_,
                                 `GTF2`_, `GFF3`_, and `PSL`_ formats
                             
    |findjuncs|                  Find all unique splice junctions in one or more
                                 transcript annotations, and optionally export these
                                 in `Tophat`_'s ``.juncs`` format
                             
    |slidejuncs|                 Compare splice junctions discovered in data to those
                                 in an annotation of known splice junctions, and,
                                 if possible with equal sequence support, slide
                                 discovered junctions to compatible known junctions.
    =========================   =============================================================================
    
"""



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.metagene`"""
import tempfile
import os
from pkg_resources import resource_filename, cleanup_resources
from nose.plugins.attrib import attr
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import (
    RPATH,
    REF_FILES,
    COUNT_OPTIONS,
    ANNOTATION_OPTIONS,
    MASK_OPTIONS,
)
from plastid.bin.test_table_equality import main as table_test
from plastid.bin.metagene import main
from plastid.util.services.decorators import catch_stderr

#===============================================================================
# INDEX: global constants used by tests
#===============================================================================

TEST_INFO = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.metagene",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="metagene"),
}

_basename = os.path.join(TEST_INFO["temp_file_path"], "test_metagene")

#===============================================================================
# INDEX: tests
#===============================================================================

tests = [
    # test generate cds start
    (
        "generate %s_cds_start --downstream 100 %s %s" %
        (_basename, ANNOTATION_OPTIONS, MASK_OPTIONS),
        [REF_FILES["yeast_metagene_cds_start"], REF_FILES["yeast_metagene_cds_start_bed"]], [
            _basename + "_cds_start_rois.txt",
            _basename + "_cds_start_rois.bed",
        ], ["", "--no_header"]
    ),
    # test generate cds stop
    (
        "generate %s_cds_stop --upstream 100 --landmark cds_stop %s %s" %
        (_basename, ANNOTATION_OPTIONS, MASK_OPTIONS), [
            REF_FILES["yeast_metagene_cds_stop"],
            REF_FILES["yeast_metagene_cds_stop_bed"],
        ], [
            _basename + "_cds_stop_rois.txt",
            _basename + "_cds_stop_rois.bed",
        ], ["", "--no_header"]
    ),
    # test count cds start with --norm_region
    (
        "count %s %s_cds_start --keep --norm_region 70 150 %s" %
        (REF_FILES["yeast_metagene_cds_start"], _basename, COUNT_OPTIONS), [
            REF_FILES["yeast_metagene_cds_start_profile"],
            REF_FILES["yeast_metagene_cds_start_normcounts"],
            REF_FILES["yeast_metagene_cds_start_rawcounts"],
        ], [
            _basename + "_cds_start_metagene_profile.txt",
            _basename + "_cds_start_normcounts.txt.gz", _basename + "_cds_start_rawcounts.txt.gz"
        ], ["", "--no_header", "--no_header"]
    ),
    # test count cds stop with --norm_region
    (
        "count %s %s_cds_stop --keep --norm_region 0 80 %s" %
        (REF_FILES["yeast_metagene_cds_stop"], _basename, COUNT_OPTIONS), [
            REF_FILES["yeast_metagene_cds_stop_profile"],
            REF_FILES["yeast_metagene_cds_stop_normcounts"],
            REF_FILES["yeast_metagene_cds_stop_rawcounts"],
        ], [
            _basename + "_cds_stop_metagene_profile.txt", _basename + "_cds_stop_normcounts.txt.gz",
            _basename + "_cds_stop_rawcounts.txt.gz"
        ], ["", "--no_header", "--no_header"]
    ),
    # test count cds start, using --normalize_over
    (
        "count %s %s_cds_start --keep --normalize_over 20 100 %s" %
        (REF_FILES["yeast_metagene_cds_start"], _basename, COUNT_OPTIONS), [
            REF_FILES["yeast_metagene_cds_start_profile"],
            REF_FILES["yeast_metagene_cds_start_normcounts"],
            REF_FILES["yeast_metagene_cds_start_rawcounts"],
        ], [
            _basename + "_cds_start_metagene_profile.txt",
            _basename + "_cds_start_normcounts.txt.gz", _basename + "_cds_start_rawcounts.txt.gz"
        ], ["", "--no_header", "--no_header"]
    ),
    # test count cds stop, using --normalize_over
    (
        "count %s %s_cds_stop --keep --normalize_over '-100' '-20' %s" %
        (REF_FILES["yeast_metagene_cds_stop"], _basename, COUNT_OPTIONS), [
            REF_FILES["yeast_metagene_cds_stop_profile"],
            REF_FILES["yeast_metagene_cds_stop_normcounts"],
            REF_FILES["yeast_metagene_cds_stop_rawcounts"],
        ], [
            _basename + "_cds_stop_metagene_profile.txt", _basename + "_cds_stop_normcounts.txt.gz",
            _basename + "_cds_stop_rawcounts.txt.gz"
        ], ["", "--no_header", "--no_header"]
    ),
]
"""Functional tests of :py:mod:`plastid.bin.metagene`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""

#===============================================================================
# INDEX: test functions
#===============================================================================


@attr(test="functional")
@attr(speed="slow")
def do_test():
    for x in execute_helper(TEST_INFO, tests):
        yield x



#!/usr/bin/env python
"""This module contains a parser for `bowtie`_'s legacy output format.
Functions in this module are are seldom used on their own, and rather are
accessed by |GenomeArray| or |SparseGenomeArray| when importing from
`bowtie`_ files.

See also
--------
|GenomeArray| and |SparseGenomeArray|
    Array-like objects that store and index quantitative data over genomes

http://bowtie-bio.sourceforge.net/manual.shtml#default-bowtie-output
    Detailed description of `bowtie`_ output format

:py:mod:`plastid.test.unit.genomics.test_genome_array`
    for integrative tests of these functions
"""
__author__ = "joshua"
__date__ = "2011-03-18"

from plastid.genomics.roitools import SegmentChain, GenomicSegment
from plastid.util.io.filters import AbstractReader

#===============================================================================
# INDEX: Readers for various bowtie1-like alignment file formats
#===============================================================================


class BowtieReader(AbstractReader):
    """Read alignments from `bowtie`_ files line-by-line into |SegmentChains|.
    The following attributes are defined and stored in the `attr` dict of
    each returned |SegmentChain|
    
    `seq_as_aligned`
        the sequence in the direction it aligns, NOT necessarily
        the read in the direction it was sequenced
    
    `qualstr_phred`
        a quality string, phred encoded
    
    `total_alignments`
        the number of total alignments found
    
    See description of `bowtie`_ legacy format at
    http://bowtie-bio.sourceforge.net/manual.shtml
    
    Parameters
    ----------
    stream : file-like
        Stream of alignments in `bowtie`_'s legacy output format
        
    
    Yields
    -------
    |SegmentChain|
        A read alignment
    """

    def filter(self, line):
        """Parse a read alignment as |SegmentChain| from a line of `bowtie`_ output"""
        items = line.strip("\n").split("\t")
        read_name = items[0]
        strand = items[1]
        ref_seq = items[2]
        coord = int(items[3])
        attr = {
            'seq_as_aligned' : items[4],
            'qualstr'        : items[5],
            'mismatch_str'   : items[7],
            'type'           : "alignment",
            'ID'             : read_name,
        } # yapf: disable

        iv = GenomicSegment(ref_seq, coord, coord + len(attr['seq_as_aligned']), strand)
        feature = SegmentChain(iv, **attr)
        return feature
